@@ -17,16 +17,20 @@ import sys
 import os
 import datetime
 import subprocess
+import tempfile
 from itertools import groupby
 from operator import itemgetter
 import pickle
+import binascii
 import logging
 import StringIO
 import re
-
+import shutil
+from hashlib import md5
 import luigi
 import luigi.hdfs
-
+import interface
+import warnings
 import mrrunner
 
 logger = logging.getLogger('luigi-interface')
@@ -39,19 +43,6 @@ def attach(*packages):
     _attached_packages.extend(packages)
 
 
-def repr_reader(inputs):
-    """Reader which uses python eval on each part of a tab separated string.
-       Yields a tuple of python objects."""
-    for input in inputs:
-        yield map(eval, input.split("\t"))
-
-
-def repr_writer(outputs, stdout):
-    """Writer which outpus the python repr for each item"""
-    for output in outputs:
-        print >> stdout, "\t".join(map(repr, output))
-
-
 def dereference(file):
     if os.path.islink(file):
         #by joining with the dirname we are certain to get the absolute path
@@ -59,6 +50,7 @@ def dereference(file):
     else:
         return file
 
+
 def get_extra_files(extra_files):
     result = []
     for f in extra_files:
@@ -82,6 +74,7 @@ def get_extra_files(extra_files):
 
     return result
 
+
 def create_packages_archive(packages, filename):
     """Create a tar archive which will contain the files for the packages listed in packages. """
     import tarfile
@@ -115,7 +108,7 @@ def create_packages_archive(packages, filename):
                     directory = '/'.join(root)
 
                     add(dereference(__import__(module_name, None, None, 'non_empty').__path__[0] + "/__init__.py"),
-                            directory + "/__init__.py")
+                        directory + "/__init__.py")
 
                 for root, dirs, files in os.walk(p):
                     if '.svn' in dirs:
@@ -149,6 +142,99 @@ def flatten(sequence):
             yield item
 
 
+class HadoopJobError(RuntimeError):
+    def __init__(self, message, out=None, err=None):
+        super(HadoopJobError, self).__init__(message, out, err)
+        self.message = message
+        self.out = out
+        self.err = err
+
+
+def run_and_track_hadoop_job(arglist):
+    ''' Runs the job by invoking the command from the given arglist. Finds tracking urls from the output and attempts to fetch
+    errors using those urls if the job fails. Throws HadoopJobError with information about the error (including stdout and stderr
+    from the process) on failure and returns normally otherwise.
+    '''
+    logger.info(' '.join(arglist))
+
+    def track_process(arglist, out_file, err_file):
+        proc = subprocess.Popen(arglist, stdout=out_file, stderr=err_file)
+        logger.info("Hadoop process output can be found at [%s] and [%s]" % (out_file.name, err_file.name))
+        proc.wait()
+        # We parse the output to try to find the tracking URL.
+        # This URL is useful for fetching the logs of the job.
+        out_file.seek(0)
+        err_file.seek(0)
+        out = out_file.read()
+        err = err_file.read()
+        tracking_url = None
+        for err_line in err.split('\n'):
+          print err_line
+          if err_line.find('Tracking URL') != -1:
+            tracking_url = err.strip().split('Tracking URL: ')[-1]
+        for out_line in out.split('\n'):
+          print out_line
+
+        if proc.returncode == 0:
+            return
+
+        # Try to fetch error logs if possible
+        message = 'Streaming job failed with exit code %d. ' % proc.returncode
+        if not tracking_url:
+            raise HadoopJobError(message + 'Also, no tracking url found.', out, err)
+
+        try:
+            task_failures = fetch_task_failures(tracking_url)
+        except Exception, e:
+            raise HadoopJobError(message + 'Additionally, an error occurred when fetching data from %s: %s' %
+                                 (tracking_url, e), out, err)
+
+        if not task_failures:
+            raise HadoopJobError(message + 'Also, could not fetch output from tasks.', out, err)
+        else:
+            raise HadoopJobError(message + 'Output from tasks below:\n%s' % task_failures, out, err)
+
+    with tempfile.NamedTemporaryFile(mode='w+r', prefix="stdout") as out_file:
+        with tempfile.NamedTemporaryFile(mode='w+r', prefix="stderr") as err_file:
+            track_process(arglist, out_file, err_file)
+
+
+def fetch_task_failures(tracking_url):
+    ''' Uses mechanize to fetch the actual task logs from the task tracker.
+
+    This is highly opportunistic, and we might not succeed. So we set a low timeout and hope it works.
+    If it does not, it's not the end of the world.
+
+    TODO: Yarn has a REST API that we should probably use instead:
+    http://hadoop.apache.org/docs/current/hadoop-yarn/hadoop-yarn-site/MapredAppMasterRest.html
+    '''
+    import mechanize
+    timeout = 3.0
+    failures_url = tracking_url.replace('jobdetails.jsp', 'jobfailures.jsp') + '&cause=failed'
+    logger.debug('Fetching data from %s' % failures_url)
+    b = mechanize.Browser()
+    b.open(failures_url, timeout=timeout)
+    links = list(b.links(text_regex='Last 4KB'))  # For some reason text_regex='All' doesn't work... no idea why
+    links = random.sample(links, min(10, len(links)))  # Fetch a random subset of all failed tasks, so not to be biased towards the early fails
+    error_text = []
+    for link in links:
+        task_url = link.url.replace('&start=-4097', '&start=-100000')  # Increase the offset
+        logger.debug('Fetching data from %s', task_url)
+        b2 = mechanize.Browser()
+        try:
+            r = b2.open(task_url, timeout=timeout)
+            data = r.read()
+        except Exception, e:
+            logger.debug('Error fetching data from %s: %s' % (task_url, e))
+            continue
+        # Try to get the hex-encoded traceback back from the output
+        for exc in re.findall(r'luigi-exc-hex=[0-9a-f]+', data):
+            error_text.append('---------- %s:' % task_url)
+            error_text.append(exc.split('=')[-1].decode('hex'))
+
+    return '\n'.join(error_text)
+
+
 class JobRunner(object):
     run_job = NotImplemented
 
@@ -167,6 +253,7 @@ class HadoopJobRunner(JobRunner):
         self.jobconfs = jobconfs
         self.input_format = input_format
         self.output_format = output_format
+        self.tmp_dir = False
 
     def run_job(self, job):
         packages = [luigi] + self.modules + job.extra_modules() + list(_attached_packages)
@@ -180,10 +267,10 @@ class HadoopJobRunner(JobRunner):
         if runner_path.endswith("pyc"):
             runner_path = runner_path[:-3] + "py"
 
-        tmp_dir = '/tmp/luigi/hadoop_job_%016x' % random.getrandbits(64)
-        # self._cleanup_tmp_dirs.append(tmp_dir)
-        logger.debug("Tmp dir: %s", tmp_dir)
-        os.makedirs(tmp_dir)
+        base_tmp_dir = interface.get_config().get('core', 'tmp-dir', '/tmp/luigi')
+        self.tmp_dir = os.path.join(base_tmp_dir, 'hadoop_job_%016x' % random.getrandbits(64))
+        logger.debug("Tmp dir: %s", self.tmp_dir)
+        os.makedirs(self.tmp_dir)
 
         # build arguments
         map_cmd = 'python mrrunner.py map'
@@ -195,14 +282,14 @@ class HadoopJobRunner(JobRunner):
         output_tmp_fn = output_final + '-temp-' + datetime.datetime.now().isoformat().replace(':', '-')
         tmp_target = luigi.hdfs.HdfsTarget(output_tmp_fn, is_tmp=True)
 
-        arglist = ['hadoop', 'jar', self.streaming_jar] + self.streaming_args
+        arglist = ['hadoop', 'jar', self.streaming_jar]
 
         # 'libjars' is a generic option, so place it first
         libjars = [libjar for libjar in self.libjars]
 
         for libjar in self.libjars_in_hdfs:
-            subprocess.call(['hadoop', 'fs', '-get', libjar, tmp_dir])
-            libjars.append(os.path.join(tmp_dir, os.path.basename(libjar)))
+            subprocess.call(['hadoop', 'fs', '-get', libjar, self.tmp_dir])
+            libjars.append(os.path.join(self.tmp_dir, os.path.basename(libjar)))
 
         if libjars:
             arglist += ['-libjars', ','.join(libjars)]
@@ -220,24 +307,20 @@ class HadoopJobRunner(JobRunner):
         if files:
             arglist += ['-files', ','.join(files)]
 
-        jobconfs = []
-
-        jobconfs.append('mapred.job.name=%s' % job.task_id)
-        jobconfs.append('mapred.reduce.tasks=%s' % job.n_reduce_tasks)
-        pool = job.pool
-        if pool is not None:
-            jobconfs.append('mapred.fairscheduler.pool=%s' % pool)
+        jobconfs = job.jobconfs()
 
         for k, v in self.jobconfs.iteritems():
             jobconfs.append('%s=%s' % (k, v))
 
         for conf in jobconfs:
-            arglist += ['-jobconf', conf]
+            arglist += ['-D', conf]
+
+        arglist += self.streaming_args
 
         arglist += ['-mapper', map_cmd, '-reducer', red_cmd]
         if job.combiner != NotImplemented:
             arglist += ['-combiner', cmb_cmd]
-        files = [runner_path, tmp_dir + '/packages.tar', tmp_dir + '/job-instance.pickle']
+        files = [runner_path, self.tmp_dir + '/packages.tar', self.tmp_dir + '/job-instance.pickle']
 
         for f in files:
             arglist += ['-file', f]
@@ -255,59 +338,23 @@ class HadoopJobRunner(JobRunner):
         arglist += ['-output', output_tmp_fn]
 
         # submit job
-        create_packages_archive(packages, tmp_dir + '/packages.tar')
+        create_packages_archive(packages, self.tmp_dir + '/packages.tar')
 
-        logger.info(' '.join(arglist))
+        job._dump(self.tmp_dir)
 
-        job._dump(tmp_dir)
-
-        proc = subprocess.Popen(arglist, stderr=subprocess.PIPE)
-
-        # We parse the output to try to find the tracking URL.
-        # This URL is useful for fetching the logs of the job.
-        tracking_url = None
-        while True:
-            line = proc.stderr.readline()
-            if not line:
-                break
-            print line.strip()
-            if line.find('Tracking URL') != -1:
-                tracking_url = line.strip().split('Tracking URL: ')[-1]
-        proc.wait()
-
-        if proc.returncode != 0:
-            # Try to fetch error logs if possible
-            if tracking_url:
-                try:
-                    self.fetch_raise_failures(tracking_url)
-                except Exception, e:
-                    raise RuntimeError('Streaming job failed with exit code %d. Additionally, an error occurred when fetching data from %s: %s' % (proc.returncode, tracking_url, e))
-
-            raise RuntimeError('Streaming job failed with exit code %d' % proc.returncode)
+        run_and_track_hadoop_job(arglist)
 
         # rename temporary work directory to given output
         tmp_target.move(output_final, fail_if_exists=True)
+        self.finish()
 
-    @staticmethod
-    def fetch_raise_failures(tracking_url):
-        ''' Uses mechanize to fetch the actual task logs from the task tracker.
+    def finish(self):
+        if self.tmp_dir and os.path.exists(self.tmp_dir):
+            logger.debug('Removing directory %s' % self.tmp_dir)
+            shutil.rmtree(self.tmp_dir)
 
-        This is highly opportunistic, and we might not succeed. So we set a low timeout and hope it works.
-        If it does not, it's not the end of the world.
-        '''
-        import mechanize
-
-        failures_url = tracking_url.replace('jobdetails.jsp', 'jobfailures.jsp')
-        b = mechanize.Browser()
-        b.open(failures_url, timeout=2.0)
-        for link in b.links(text_regex='Last 4KB'):
-            b2 = mechanize.Browser()
-            r = b2.open(link.url, timeout=1.0)
-            data = r.read()
-            # Try to get the hex-encoded traceback back from the output
-            for exc in re.findall(r'luigi-exc-hex=[0-9a-f]+', data):
-                print '---------- %s:' % link.url
-                print exc.split('=')[-1].decode('hex')
+    def __del__(self):
+        self.finish()
 
 
 class DefaultHadoopJobRunner(HadoopJobRunner):
@@ -339,10 +386,11 @@ class LocalJobRunner(JobRunner):
     def group(self, input):
         output = StringIO.StringIO()
         lines = []
-        for line in input:
-            k, v = line.strip().split('\t')
-            lines.append((k, line))
-        for k, line in sorted(lines):
+        for i, line in enumerate(input):
+            parts = line.rstrip('\n').split('\t')
+            blob = md5(str(i)).hexdigest()  # pseudo-random blob to make sure the input isn't sorted
+            lines.append((parts[:-1], blob, line))
+        for k, _, line in sorted(lines):
             output.write(line)
         output.seek(0)
         return output
@@ -374,9 +422,28 @@ class LocalJobRunner(JobRunner):
         reduce_output.close()
 
 
-class JobTask(luigi.Task):
+class BaseHadoopJobTask(luigi.Task):
     n_reduce_tasks = 25
     pool = luigi.Parameter(is_global=True, default=None, significant=False)
+    # This value can be set to change the default batching increment. Default is 1 for backwards compatibility.
+    batch_counter_default = 1
+
+    final_mapper = NotImplemented
+    final_combiner = NotImplemented
+    final_reducer = NotImplemented
+
+    _counter_dict = {}
+    task_id = None
+
+    def jobconfs(self):
+        jcs = []
+        jcs.append('mapred.job.name=%s' % self.task_id)
+        jcs.append('mapred.reduce.tasks=%s' % self.n_reduce_tasks)
+        pool = self.pool
+        if pool is not None:
+            jcs.append('mapred.fairscheduler.pool=%s' % pool)
+        return jcs
+
 
     def init_local(self):
         ''' Implement any work to setup any internal datastructure etc here.
@@ -390,22 +457,10 @@ class JobTask(luigi.Task):
     def init_hadoop(self):
         pass
 
-    def init_mapper(self):
-        pass
-
-    def init_combiner(self):
-        pass
-
-    def init_reducer(self):
-        pass
-
     def run(self):
         self.init_local()
         self.job_runner().run_job(self)
 
-    def _setup_remote(self):
-        self._setup_links()
-
     def requires_local(self):
         ''' Default impl - override this method if you need any local input to be accessible in init() '''
         return []
@@ -423,9 +478,50 @@ class JobTask(luigi.Task):
         # Overrides the default implementation
         return luigi.task.flatten(self.requires_hadoop()) + luigi.task.flatten(self.requires_local())
 
+    def on_failure(self, exception):
+        if isinstance(exception, HadoopJobError):
+            return """Hadoop job failed with message: {message}
+
+    stdout:
+    {stdout}
+
+
+    stderr:
+    {stderr}
+      """.format(message=exception.message, stdout=exception.out, stderr=exception.err)
+        else:
+            return super(BaseHadoopJobTask, self).on_failure(exception)
+
+
+
+class JobTask(BaseHadoopJobTask):
+
+    def init_mapper(self):
+        pass
+
+    def init_combiner(self):
+        pass
+
+    def init_reducer(self):
+        pass
+
+    def _setup_remote(self):
+        self._setup_links()
+
     def job_runner(self):
         # We recommend that you define a subclass, override this method and set up your own config
-        return DefaultHadoopJobRunner()
+        """ Get the MapReduce runner for this job
+
+        If all outputs are HdfsTargets, the DefaultHadoopJobRunner will be used. Otherwise, the LocalJobRunner which streams all data through the local machine will be used (great for testing).
+        """
+        outputs = luigi.task.flatten(self.output())
+        for output in outputs:
+            if not isinstance(output, luigi.hdfs.HdfsTarget):
+                warnings.warn("Job is using one or more non-HdfsTarget outputs" +
+                              " so it will be run in local mode")
+                return LocalJobRunner()
+        else:
+            return DefaultHadoopJobRunner()
 
     def reader(self, input_stream):
         """Reader is a method which iterates over input lines and outputs records.
@@ -456,7 +552,38 @@ class JobTask(luigi.Task):
         for v in values:
             yield key, v
 
-    def incr_counter(self, *args):
+    def incr_counter(self, *args, **kwargs):
+        """ Increments a Hadoop counter
+
+        Since counters can be a bit slow to update, this batches the updates.
+        """
+        threshold = kwargs.get("threshold", self.batch_counter_default)
+        if len(args) == 2:
+            # backwards compatibility with existing hadoop jobs
+            group_name, count = args
+            key = (group_name,)
+        else:
+            group, name, count = args
+            key = (group, name)
+
+        ct = self._counter_dict.get(key, 0)
+        ct += count
+        if ct >= threshold:
+            new_arg = list(key)+[ct]
+            self._incr_counter(*new_arg)
+            ct = 0
+        self._counter_dict[key] = ct
+
+    def _flush_batch_incr_counter(self):
+        """ Increments any unflushed counter values
+        """
+        for key, count in self._counter_dict.iteritems():
+            if count == 0:
+                continue
+            args = list(key) + [count]
+            self._incr_counter(*args)
+
+    def _incr_counter(self, *args):
         """ Increments a Hadoop counter
 
         Note that this seems to be a bit slow, ~1 ms. Don't overuse this function by updating very frequently.
@@ -492,7 +619,10 @@ class JobTask(luigi.Task):
                 d = os.path.dirname(dst)
                 if d and not os.path.exists(d):
                     os.makedirs(d)
-                os.link(src, dst)
+                if not os.path.exists(dst):
+                    # If the combiner runs, the file might already exist,
+                    # so no reason to create the link again
+                    os.link(src, dst)
 
     def _dump(self, dir=''):
         """Dump instance to file."""
@@ -515,32 +645,63 @@ class JobTask(luigi.Task):
         for record in self.reader(input_stream):
             for output in self.mapper(*record):
                 yield output
+        if self.final_mapper != NotImplemented:
+            for output in self.final_mapper():
+                yield output
+        self._flush_batch_incr_counter()
 
-    def _reduce_input(self, inputs, reducer):
+    def _reduce_input(self, inputs, reducer, final=NotImplemented):
         """Iterate over input, collect values with the same key, and call the reducer for each uniqe key."""
         for key, values in groupby(inputs, itemgetter(0)):
             for output in reducer(key, (v[1] for v in values)):
                 yield output
+        if final != NotImplemented:
+            for output in final():
+                yield output
+        self._flush_batch_incr_counter()
 
     def _run_mapper(self, stdin=sys.stdin, stdout=sys.stdout):
         """Run the mapper on the hadoop node."""
         self.init_hadoop()
         self.init_mapper()
         outputs = self._map_input((line[:-1] for line in stdin))
-        repr_writer(outputs, stdout)
+        self.internal_writer(outputs, stdout)
 
     def _run_reducer(self, stdin=sys.stdin, stdout=sys.stdout):
         """Run the reducer on the hadoop node."""
         self.init_hadoop()
         self.init_reducer()
-        outputs = self._reduce_input(repr_reader((line[:-1] for line in stdin)), reducer=self.reducer)
+        outputs = self._reduce_input(self.internal_reader((line[:-1] for line in stdin)), self.reducer, self.final_reducer)
         self.writer(outputs, stdout)
 
     def _run_combiner(self, stdin=sys.stdin, stdout=sys.stdout):
         self.init_hadoop()
         self.init_combiner()
-        outputs = self._reduce_input(repr_reader((line[:-1] for line in stdin)), reducer=self.combiner)
-        repr_writer(outputs, stdout)
+        outputs = self._reduce_input(self.internal_reader((line[:-1] for line in stdin)), self.combiner, self.final_combiner)
+        self.internal_writer(outputs, stdout)
 
-    def _print_exception(self, exc):
-        print >> sys.stderr, 'luigi-exc-hex=%s' % exc.encode('hex')
+    def internal_reader(self, input_stream):
+        """Reader which uses python eval on each part of a tab separated string.
+        Yields a tuple of python objects."""
+        for input in input_stream:
+            yield map(eval, input.split("\t"))
+
+    def internal_writer(self, outputs, stdout):
+        """Writer which outputs the python repr for each item"""
+        for output in outputs:
+            print >> stdout, "\t".join(map(repr, output))
+
+
+def pickle_reader(job, input_stream):
+    def decode(item):
+        return pickle.loads(binascii.a2b_base64(item))
+    for line in input_stream:
+        items = line.split('\t')
+        yield map(decode, items)
+
+
+def pickle_writer(job, outputs, stdout):
+    def encode(item):
+        return binascii.b2a_base64(pickle.dumps(item))[:-1]  # remove trailing newline
+    for keyval in outputs:
+        print >> stdout, "\t".join(map(encode, keyval))
